@@ -12,6 +12,7 @@ from datetime import datetime, timedelta
 from plugwise.constants import (
     ACK_ERROR,
     ACK_TIMEOUT,
+    CB_NEW_NODE,
     MAX_TIME_DRIFT,
     MESSAGE_TIME_OUT,
     MESSAGE_RETRY,
@@ -76,22 +77,27 @@ class stick(object):
     """
 
     def __init__(self, port, callback=None, print_progress=False):
-        self.logger = logging.getLogger("plugwise")
+        self.logger = logging.getLogger("python-plugwise")
         self._mac_stick = None
         self.port = port
         self.network_online = False
         self.circle_plus_mac = None
+        self._circle_plus_discovered = False
+        self._circle_plus_retries = 0
         self.network_id = None
         self.parser = PlugwiseParser(self)
         self._plugwise_nodes = {}
-        self._nodes_to_discover = []
+        self._nodes_registered = 0
+        self._nodes_to_discover = {}
+        self._nodes_not_discovered = {}
+        self._stick_initialized = False
+        self._stick_callbacks = {}
         self.last_ack_seq_id = None
         self.expected_responses = {}
         self.print_progress = print_progress
         self.timezone_delta = datetime.now().replace(
             minute=0, second=0, microsecond=0
         ) - datetime.utcnow().replace(minute=0, second=0, microsecond=0)
-        
         self._run_receive_timeout_thread = False
         self._run_send_message_thread = False
         self._run_update_thread = False
@@ -126,7 +132,7 @@ class stick(object):
         except TimeoutException as e:
             self.logger.error("Timeout exception while initializing USBstick")
         except Exception as e:
-            self.logger.error("Unknown error : %s"), e
+            self.logger.error("Unknown error : %s", e)
 
     def connect(self, callback=None):
         """ Connect to stick and raise error if it fails"""
@@ -138,13 +144,13 @@ class stick(object):
         else:
             self.logger.debug("Open USB serial connection to Plugwise Zigbee stick")
             self.connection = PlugwiseUSBConnection(self.port, self)
-        self.connection.open_port()
+        self.connection.connect()
 
         self.logger.debug("Starting threads...")
         # receive timeout deamon
         self._run_receive_timeout_thread = True
         self._receive_timeout_thread = threading.Thread(
-            None, self._receive_timeout_loop, "receive_timeout_deamon", (), {}
+            None, self._receive_timeout_loop, "receive_timeout_thread", (), {}
         )
         self._receive_timeout_thread.daemon = True
         self._receive_timeout_thread.start()
@@ -152,7 +158,7 @@ class stick(object):
         self._send_message_queue = Queue()
         self._run_send_message_thread = True
         self._send_message_thread = threading.Thread(
-            None, self._send_message_loop, "send_messages_deamon", (), {}
+            None, self._send_message_loop, "send_messages_thread", (), {}
         )
         self._send_message_thread.daemon = True
         self._send_message_thread.start()
@@ -160,51 +166,106 @@ class stick(object):
         self._run_update_thread = False
         self._auto_update_timer = None
         self._update_thread = threading.Thread(
-            None, self._update_loop, "update_daemon", (), {}
+            None, self._update_loop, "update_thread", (), {}
         )
         self._update_thread.daemon = True
         self.logger.debug("All threads started")
 
-    def initialize_stick(self, callback=None) -> bool:
+    def initialize_stick(self, callback=None, timeout=MESSAGE_TIME_OUT):
         # Initialize USBstick
         if not self.connection.is_connected():
             raise StickInitError
-        self.init_finished = False
 
         def cb_stick_initialized():
             """ Callback when initialization of Plugwise USBstick is finished """
-            self.init_finished = True
-            if not self.network_online:
-                raise NetworkDown
+            self._stick_initialized = True
 
             # Start watchdog deamon
             self._run_watchdog = True
             self._watchdog_thread = threading.Thread(
-                None, self._watchdog_loop, "watchdog_daemon", (), {}
+                None, self._watchdog_loop, "watchdog_thread", (), {}
             )
             self._watchdog_thread.daemon = True
             self._watchdog_thread.start()
 
+            # Try to discover Circle+
+            if self.circle_plus_mac:
+                self.discover_node(self.circle_plus_mac)
             if callback:
                 callback()
 
         self.logger.debug("Send init request to Plugwise Zigbee stick")
         self.send(StickInitRequest(), cb_stick_initialized)
-        timeout = 0
-        while not self.init_finished and (timeout < MESSAGE_TIME_OUT):
-            timeout += 1
-            time.sleep(1)
-        if not self.init_finished:
+        time_counter = 0
+        while not self._stick_initialized and (time_counter < timeout):
+            time_counter += 0.1
+            time.sleep(0.1)
+        if not self._stick_initialized:
             raise StickInitError
+        if not self.network_online:
+            raise NetworkDown
+
+    def initialize_circle_plus(self, callback=None, timeout=MESSAGE_TIME_OUT):
+        # Initialize Circle+
+        if (
+            not self.connection.is_connected()
+            or not self._stick_initialized
+            or not self.circle_plus_mac
+        ):
+            raise StickInitError
+        # discover circle+ node
+        self.discover_node(self.circle_plus_mac)
+
+        time_counter = 0
+        while not self._circle_plus_discovered and (time_counter < timeout):
+            time_counter += 0.1
+            time.sleep(0.1)
+        if not self._circle_plus_discovered:
+            raise CirclePlusError
 
     def disconnect(self):
         """ Disconnect from stick and raise error if it fails"""
-        try:
-            self.connection.close_port()
-        except Exception as e:
-            self.logger.error(
-                "Error while disconnect port: %s", e,
-            )
+        self._run_watchdog = False
+        self._run_update_thread = False
+        self._auto_update_timer = None
+        self._run_send_message_thread = False
+        self._run_receive_timeout_thread = False
+        self.connection.disconnect()
+
+    def subscribe_stick_callback(self, callback, callback_type):
+        """ Subscribe callback to execute """
+        if callback_type not in self._stick_callbacks:
+            self._stick_callbacks[callback_type] = []
+        self._stick_callbacks[callback_type].append(callback)
+
+    def unsubscribe_stick_callback(self, callback, callback_type):
+        """ Register callback to execute """
+        if callback_type in self._stick_callbacks:
+            self._stick_callbacks[callback_type].remove(callback)
+
+    def do_callback(self, callback_type, callback_arg=None):
+        """ Execute callbacks registered for specified callback type """
+        if callback_type in self._stick_callbacks:
+            for callback in self._stick_callbacks[callback_type]:
+                try:
+                    callback(callback_arg)
+                except Exception as e:
+                    self.logger.error("Error while executing callback : %s", e)
+
+    def _discover_after_scan(self):
+        """ Helper to do callback for new node """
+        mac_found = None
+        for mac in self._nodes_not_discovered.keys():
+            if mac in self._plugwise_nodes:
+                self.do_callback(CB_NEW_NODE, mac)
+                mac_found = mac
+        if mac_found:
+            del self._nodes_not_discovered[mac_found]
+
+    def registered_nodes(self) -> int:
+        """ Return number of nodes registered in Circle+ """
+        # Include Circle+ too
+        return self._nodes_registered + 1
 
     def nodes(self) -> list:
         """ Return mac addresses of known plugwise nodes """
@@ -217,15 +278,23 @@ class stick(object):
             return self._plugwise_nodes[mac]
         return None
 
-    def discover_node(self, mac, callback=None) -> bool:
-        """ Discovery plugwise node """
-        assert isinstance(mac, str)
+    def discover_node(self, mac: str, callback=None) -> bool:
+        """ Discovery of plugwise node """
         if validate_mac(mac) == True:
-            self.send(
-                NodeInfoRequest(bytes(mac, "ascii")), callback,
-            )
-            return True
-        return False
+            if mac not in self._plugwise_nodes.keys():
+                if mac not in self._nodes_not_discovered.keys():
+                    self._nodes_not_discovered[mac] = (
+                        None,
+                        None,
+                    )
+                self.send(
+                    NodeInfoRequest(bytes(mac, "ascii")), callback,
+                )
+                return True
+            else:
+                return False
+        else:
+            return False
 
     def scan(self, callback=None):
         """ scan for connected plugwise nodes """
@@ -236,6 +305,7 @@ class stick(object):
             self.logger.debug("Scan plugwise network finished")
             self._nodes_discovered = 0
             self._nodes_to_discover = nodes_to_discover
+            self._nodes_registered = len(nodes_to_discover)
             self._discovery_finished = False
 
             def node_discovered():
@@ -243,25 +313,23 @@ class stick(object):
                 self.logger.debug(
                     "Discovered Plugwise node %s of %s",
                     str(len(self._plugwise_nodes)),
-                    str(self._nodes_to_discover),
+                    str(len(self._nodes_to_discover)),
                 )
                 if (len(self._plugwise_nodes) - 1) >= len(self._nodes_to_discover):
                     self._discovery_finished = True
-                    self._nodes_to_discover = None
-                    if callback != None:
+                    self._nodes_to_discover = {}
+                    if callback:
                         callback()
 
             def timeout_expired():
                 if not self._discovery_finished:
-                    for (mac, address) in self._nodes_to_discover:
+                    for mac in self._nodes_to_discover:
                         if mac not in self._plugwise_nodes.keys():
                             self.logger.warning(
-                                "Failed to discover Plugwise node %s before timeout expired.",
+                                "Failed to discover registered Plugwise node with MAC '%s' before timeout expired.",
                                 str(mac),
                             )
-                            # do 1 retry
-                            self.send(NodeInfoRequest(bytes(mac, "ascii")))
-                    if callback != None:
+                    if callback:
                         callback()
 
             # setup timeout for loading nodes
@@ -272,10 +340,8 @@ class stick(object):
                 discover_timeout, timeout_expired
             ).start()
             self.logger.debug("Start discovery of linked node types...")
-            for (mac, address) in nodes_to_discover:
-                self.send(
-                    NodeInfoRequest(bytes(mac, "ascii")), node_discovered,
-                )
+            for mac in nodes_to_discover:
+                self.discover_node(mac, node_discovered)
 
         def scan_circle_plus():
             """Callback when Circle+ is discovered"""
@@ -285,14 +351,22 @@ class stick(object):
                 self.logger.debug("Scan Circle+ for linked nodes...")
                 self._plugwise_nodes[self.circle_plus_mac].scan_for_nodes(scan_finished)
             else:
-                self.logger.error("Circle+ is not discovered in %s", self._plugwise_nodes)
+                self.logger.error(
+                    "Circle+ is not discovered in %s", self._plugwise_nodes
+                )
 
         # Discover Circle+
         if self.circle_plus_mac:
-            if self.print_progress:
-                print("Discover Circle+")
-            self.logger.debug("Discover Circle+ at %s", self.circle_plus_mac)
-            self.discover_node(self.circle_plus_mac, scan_circle_plus)
+            if self.circle_plus_mac in self._plugwise_nodes:
+                if self.print_progress:
+                    print("Scan Circle+ for linked nodes")
+                self.logger.debug("Scan Circle+ for linked nodes...")
+                self._plugwise_nodes[self.circle_plus_mac].scan_for_nodes(scan_finished)
+            else:
+                if self.print_progress:
+                    print("Discover Circle+")
+                self.logger.debug("Discover Circle+ at %s", self.circle_plus_mac)
+                self.discover_node(self.circle_plus_mac, scan_circle_plus)
         else:
             self.logger.error(
                 "Plugwise stick not properly initialized, Circle+ MAC is missing."
@@ -365,7 +439,7 @@ class stick(object):
         """ deamon to send messages in queue """
         while self._run_send_message_thread:
             request_set = self._send_message_queue.get(block=True)
-            if self.last_ack_seq_id != None:
+            if self.last_ack_seq_id:
                 # Calc new seq_id based last received ack messsage
                 seq_id = inc_seq_id(self.last_ack_seq_id)
             else:
@@ -406,7 +480,7 @@ class stick(object):
             ):
                 time.sleep(0.1)
                 timeout_counter += 1
-            if timeout_counter > 10:
+            if timeout_counter > 10 and self._run_send_message_thread:
                 if seq_id in self.expected_responses:
                     if self.expected_responses[seq_id][3] <= MESSAGE_RETRY:
                         self.logger.warning(
@@ -513,11 +587,12 @@ class stick(object):
         elif isinstance(message, NodeInfoResponse):
             if not mac in self._plugwise_nodes:
                 if message.node_type.value == NODE_TYPE_CIRCLE_PLUS:
+                    self._circle_plus_discovered = True
                     self._append_node(mac, 0, message.node_type.value)
                 else:
-                    for (mac_to_discover, address) in self._nodes_to_discover:
+                    for mac_to_discover in self._nodes_to_discover:
                         if mac == mac_to_discover:
-                            self._append_node(mac, address, message.node_type.value)
+                            self._append_node(mac, self._nodes_to_discover[mac_to_discover], message.node_type.value)
             self._plugwise_nodes[mac].on_message(message)
         else:
             if mac in self._plugwise_nodes:
@@ -606,21 +681,12 @@ class stick(object):
                             )
             del self.expected_responses[seq_id]
 
-    def stop(self):
-        """
-        Stop connection to Plugwise Zigbee network
-        """
-        self._run_watchdog = False
-        self._auto_update_timer = None
-        self._run_receive_timeout_thread = False
-        self._run_send_message_thread = False
-        self.connection.stop_connection()
-
     def _watchdog_loop(self):
         """
         Main worker loop to watch all other worker threads
         """
         time.sleep(5)
+        circle_plus_retry_counter = 0
         while self._run_watchdog:
             # Connection
             if self.connection.is_connected():
@@ -637,7 +703,11 @@ class stick(object):
                         "Unexpected halt of receive thread, restart thread",
                     )
                     self._receive_timeout_thread = threading.Thread(
-                        None, self._receive_timeout_loop, "receive_timeout_deamon", (), {}
+                        None,
+                        self._receive_timeout_loop,
+                        "receive_timeout_thread",
+                        (),
+                        {},
                     )
                     self._receive_timeout_thread.daemon = True
                     self._receive_timeout_thread.start()
@@ -648,7 +718,7 @@ class stick(object):
                         "Unexpected halt of send thread, restart thread",
                     )
                     self._send_message_thread = threading.Thread(
-                        None, self._send_message_loop, "send_messages_deamon", (), {}
+                        None, self._send_message_loop, "send_messages_thread", (), {}
                     )
                     self._send_message_thread.daemon = True
                     self._send_message_thread.start()
@@ -660,10 +730,21 @@ class stick(object):
                     )
                     self._run_update_thread = True
                     self._update_thread = threading.Thread(
-                        None, self._update_loop, "update_daemon", (), {}
+                        None, self._update_loop, "update_thread", (), {}
                     )
                     self._update_thread.daemon = True
                     self._update_thread.start()
+            # Circle+ discovery
+            if self._circle_plus_discovered == False:
+                # First hour every once an hour
+                if self._circle_plus_retries < 60 or circle_plus_retry_counter > 60:
+                    self.logger.debug(
+                        "Circle+ not yet discovered, resubmit discovery request",
+                    )
+                    self.discover_node(self.circle_plus_mac, self.scan)
+                    self._circle_plus_retries += 1
+                    circle_plus_retry_counter = 0
+                circle_plus_retry_counter += 1
             time.sleep(WATCHDOG_DEAMON)
 
     def _update_loop(self):
@@ -673,6 +754,7 @@ class stick(object):
         """
         self._run_update_thread = True
         self._auto_update_first_run = True
+        day_of_month = datetime.now().day
         try:
             while self._run_update_thread:
                 for mac in self._plugwise_nodes:
@@ -682,11 +764,13 @@ class stick(object):
                     )
                     self._plugwise_nodes[mac].ping()
                     # Only power use updates for supported nodes
-                    if isinstance(self._plugwise_nodes[mac], PlugwiseCircle) or isinstance(
-                        self._plugwise_nodes[mac], PlugwiseCirclePlus
-                    ):
+                    if isinstance(
+                        self._plugwise_nodes[mac], PlugwiseCircle
+                    ) or isinstance(self._plugwise_nodes[mac], PlugwiseCirclePlus):
                         # Don't check at first time
-                        self.logger.debug("Request current power usage for node %s", mac)
+                        self.logger.debug(
+                            "Request current power usage for node %s", mac
+                        )
                         if not self._auto_update_first_run and self._run_update_thread:
                             # Only request update if node is available
                             if self._plugwise_nodes[mac].get_available():
@@ -713,26 +797,65 @@ class stick(object):
                                 if self._plugwise_nodes[mac]._last_info_message != None:
                                     if self._plugwise_nodes[mac]._last_info_message < (
                                         datetime.now().replace(
-                                            minute=1, second=MAX_TIME_DRIFT, microsecond=0
+                                            minute=1,
+                                            second=MAX_TIME_DRIFT,
+                                            microsecond=0,
                                         )
                                     ):
                                         self._plugwise_nodes[mac]._request_info(
-                                            self._plugwise_nodes[mac]._request_power_buffer
+                                            self._plugwise_nodes[
+                                                mac
+                                            ]._request_power_buffer
                                         )
                                 if not self._plugwise_nodes[mac]._last_log_collected:
                                     self._plugwise_nodes[mac]._request_power_buffer()
                         else:
                             if self._run_update_thread:
                                 self.logger.debug(
-                                    "First request for current power usage for node %s", mac
+                                    "First request for current power usage for node %s",
+                                    mac,
                                 )
                                 self._plugwise_nodes[mac].update_power_usage()
                 self._auto_update_first_run = False
+
+                # Try to rediscover node(s) which where not available at initial scan
+                # Do this the first hour at every update, there after only once an hour
+                for mac in self._nodes_not_discovered:
+                    (firstrequest, lastrequest) = self._nodes_not_discovered[mac]
+                    if firstrequest and lastrequest:
+                        if (firstrequest + timedelta(hours=1)) > datetime.now():
+                            # first hour, so do every update a request
+                            self.discover_node(mac, self._discover_after_scan)
+                            self._nodes_not_discovered[mac] = (
+                                firstrequest,
+                                datetime.now(),
+                            )
+                        else:
+                            if (lastrequest + timedelta(hours=1)) < datetime.now():
+                                self.discover_node(mac, self._discover_after_scan)
+                                self._nodes_not_discovered[mac] = (
+                                    firstrequest,
+                                    datetime.now(),
+                                )
+                    else:
+                        self.discover_node(mac, self._discover_after_scan)
+                        self._nodes_not_discovered[mac] = (
+                            datetime.now(),
+                            datetime.now(),
+                        )
+                # Sync internal clock of all available nodes once a day
+                if datetime.now().day != day_of_month:
+                    day_of_month = datetime.now().day
+                    for mac in self._plugwise_nodes:
+                        if self._plugwise_nodes[mac].get_available():
+                            self._plugwise_nodes[mac].sync_clock()
                 if self._auto_update_timer:
                     time.sleep(self._auto_update_timer)
         except Exception as e:
             exc_type, exc_obj, exc_tb = sys.exc_info()
-            self.logger.error("Error at line %s of _update_loop : %s", exc_tb.tb_lineno, e)
+            self.logger.error(
+                "Error at line %s of _update_loop : %s", exc_tb.tb_lineno, e
+            )
 
     def auto_update(self, timer=None):
         """
